# -*- encoding: utf-8 -*-
from __future__ import print_function, unicode_literals, division
import logging

import threading
try:
    import queue
except ImportError:
    import Queue as queue
from enocean.protocol.packet import Packet
from enocean.protocol.constants import PARSE_RESULT

logger = logging.getLogger('enocean.communicators.Communicator')


class Communicator(threading.Thread):
    '''
    Communicator base-class for EnOcean.
    Not to be used directly, only serves as base class for SerialCommunicator etc.
    '''
    def __init__(self):
        super(Communicator, self).__init__()
        # Create an event to stop the thread
        self._stop_flag = threading.Event()
        # Input buffer
        self._buffer = []
        # Setup packet queues
        self.transmit = queue.Queue()
        self.receive = queue.Queue()

    def _get_from_send_queue(self):
        ''' Get message from send queue, if one exists '''
        try:
            p = self.transmit.get(block=False)
            logger.info('Sending packet')
            logger.debug(p)
            return p
        except queue.Empty:
            pass
        return None

    def send(self, packet):
        if not isinstance(packet, Packet):
            logger.error('Object to send must be an instance of Packet')
            return False
        self.transmit.put(packet)
        return True

    def stop(self):
        self._stop_flag.set()

    def parse(self):
        ''' Parses messages and puts them to receive queue '''
        # Loop while we get new messages
        while True:
            status, self._buffer, p = Packet.parse_msg(self._buffer)
            # If message is incomplete -> break the loop
            if status == PARSE_RESULT.INCOMPLETE:
                return status

            # If message is OK, add it to receive queue
            if status == PARSE_RESULT.OK and p:
                self.receive.put(p)
                logger.debug(p)



# -*- encoding: utf-8 -*-
from __future__ import print_function, unicode_literals, division
import logging
import serial

from enocean.communicators.communicator import Communicator

logger = logging.getLogger('enocean.communicators.SerialCommunicator')


class SerialCommunicator(Communicator):
    ''' Serial port communicator class for EnOcean radio '''
    def __init__(self, port='/dev/ttyAMA0'):
        super(SerialCommunicator, self).__init__()
        # Initialize serial port
        self.__ser = serial.Serial(port, 57600, timeout=0.1)

    def run(self):
        logger.info('SerialCommunicator started')
        while not self._stop_flag.is_set():
            # If there's messages in transmit queue
            # send them
            while True:
                p = self._get_from_send_queue()
                if not p:
                    break
                self.__ser.write(bytearray(p.build()))

            # Read chars from serial port as hex numbers
            try:
                self._buffer.extend(bytearray(self.__ser.read(16)))
            except serial.SerialException:
                logger.error('Serial port exception! (device disconnected or multiple access on port?)')
                break
            self.parse()

        self.__ser.close()
        logger.info('SerialCommunicator stopped')



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
from enocean.consolelogger import init_logging
from enocean.communicators.serialcommunicator import SerialCommunicator
from enocean.protocol.packet import Packet
from enocean.protocol.constants import PACKET, RORG
import sys
import traceback

try:
    import queue
except ImportError:
    import Queue as queue

p = Packet(PACKET.COMMON_COMMAND, [0x08])

init_logging()
c = SerialCommunicator()
c.start()
c.send(p)
while c.is_alive():
    try:
        # Loop to empty the queue...
        p = c.receive.get(block=True, timeout=1)
        if p.type == PACKET.RADIO and p.rorg == RORG.BS4:
            for k in p.parse_eep(0x02, 0x05):
                print('%s: %s' % (k, p.parsed[k]))
        if p.type == PACKET.RADIO and p.rorg == RORG.BS1:
            for k in p.parse_eep(0x00, 0x01):
                print('%s: %s' % (k, p.parsed[k]))
        if p.type == PACKET.RADIO and p.rorg == RORG.RPS:
            for k in p.parse_eep(0x02, 0x04):
                print('%s: %s' % (k, p.parsed[k]))
    except queue.Empty:
        continue
    except KeyboardInterrupt:
        break
    except Exception:
        traceback.print_exc(file=sys.stdout)
        break

if c.is_alive():
    c.stop()



# -*- encoding: utf-8 -*-
from __future__ import print_function, unicode_literals, division
import os
from bs4 import BeautifulSoup

path = os.path.dirname(os.path.realpath(__file__))


class EEP(object):
    def __init__(self):
        self.ok = False
        try:
            with open(os.path.join(path, 'EEP_2.6.1.xml'), 'r') as f:
                self.soup = BeautifulSoup(f.read())
            self.ok = True
        except IOError:
            pass

    def _get_hex(self, nmb):
        ''' Get hex-representation of number '''
        return '0x%02X' % (nmb)

    def _get_raw(self, value, bitarray):
        ''' Get raw data as integer, based on offset and size '''
        offset = int(value['offset'])
        size = int(value['size'])
        return int(''.join(['1' if digit else '0' for digit in bitarray[offset:offset + size]]), 2)

    def _get_value(self, value, data):
        ''' Get value, based on the data in XML '''
        raw_value = self._get_raw(value, data)

        rng = value.find('range')
        rng_min = float(rng.find('min').text)
        rng_max = float(rng.find('max').text)

        scl = value.find('scale')
        scl_min = float(scl.find('min').text)
        scl_max = float(scl.find('max').text)

        return {
            value['shortcut']: {
                'description': value['description'],
                'unit': value['unit'],
                'value': (scl_max - scl_min) / (rng_max - rng_min) * (raw_value - rng_min) + scl_min,
                'raw_value': raw_value,
            }
        }

    def _get_enum(self, value, data):
        ''' Get enum value, based on the data in XML '''
        raw_value = self._get_raw(value, data)
        value_desc = value.find('item', {'value': str(raw_value)})
        return {
            value['shortcut']: {
                'description': value['description'],
                'unit': value.get('unit', ''),
                'value': value_desc['description'],
                'raw_value': raw_value,
            }
        }

    def find_profile(self, rorg, func, type):
        ''' Find profile and data description, matching RORG, FUNC and TYPE '''
        if not self.ok:
            return None, None

        rorg = self.soup.find('telegram', {'rorg': self._get_hex(rorg)})
        if not rorg:
            return None, None

        func = rorg.find('profiles', {'func': self._get_hex(func)})
        if not func:
            return None, None

        profile = rorg.find('profile', {'type': self._get_hex(type)})
        data_description = profile.find('data')
        if not data_description:
            data_description = []

        return profile, data_description

    def get_values(self, rorg, func, type, data):
        ''' Get keys and values from data, matching RORG, FUNC and TYPE '''
        if not self.ok:
            return [], {}

        profile, data_description = self.find_profile(rorg, func, type)
        if not profile or not data_description:
            return [], {}

        output = {}
        for d in data_description.contents:
            if not d.name:
                continue
            if d.name == 'value':
                output.update(self._get_value(d, data))
            if d.name == 'enum':
                output.update(self._get_enum(d, data))
        return output.keys(), output



# -*- encoding: utf-8 -*-
from __future__ import print_function, unicode_literals, division
import logging

from enocean.protocol import crc8
from enocean.protocol.eep import EEP
from enocean.protocol.constants import PACKET, RORG, PARSE_RESULT

logger = logging.getLogger('enocean.protocol.packet')
eep = EEP()


class Packet(object):
    '''
    Base class for Packet.
    Mainly used for for packet generation and
    Packet.parse_msg(buf) for parsing message.
    parse_msg() returns subclass, if one is defined for the data type.
    '''
    def __init__(self, type, data=[], optional=[]):
        self.type = type
        self.rorg = RORG.UNDEFINED
        self.data = data
        self.optional = optional
        self.bit_data = []
        self.bit_optional = []
        self.parsed = {}

        self.parse()

    def __str__(self):
        return '0x%02X %s %s %s' % (self.type, [hex(o) for o in self.data], [hex(o) for o in self.optional], self.parsed)

    def __unicode__(self):
        return self.__str__()

    def __getattr__(self, attr):
        if attr in self.parsed:
            return self.parsed[attr]
        raise AttributeError

    def __eq__(self, other):
        return self.type == other.type and self.rorg == other.rorg and self.data == other.data and self.optional == other.optional

    def _get_bit(self, byte, bit):
        ''' Get bit value from byte '''
        return (byte >> bit) & 0x01

    def _combine_hex(self, data):
        ''' Combine list of integer values to one big integer '''
        output = 0x00
        for i, d in enumerate(reversed(data)):
            output |= (d << i * 8)
        return output

    def _to_bitarray(self, data, width=8):
        ''' Convert data (list of integers or integer) to bitarray '''
        if isinstance(data, list):
            data = self._combine_hex(data)
        return [True if digit == '1' else False for digit in bin(data)[2:].zfill(width)]

    @staticmethod
    def parse_msg(buf):
        '''
        Parses message from buffer.
        returns:
            - PARSE_RESULT
            - remaining buffer
            - Packet -object (if message was valid, else None)
        '''
        # If the buffer doesn't contain 0x55 (start char)
        # the message isn't needed -> ignore
        if 0x55 not in buf:
            return PARSE_RESULT.INCOMPLETE, [], None

        # Valid buffer starts from 0x55
        buf = buf[buf.index(0x55):]
        try:
            data_len = (buf[1] << 8) | buf[2]
            opt_len = buf[3]
        except IndexError:
            # If the fields don't exist, message is incomplete
            return PARSE_RESULT.INCOMPLETE, buf, None

        # Header: 6 bytes, data, optional data and data checksum
        msg_len = 6 + data_len + opt_len + 1
        if len(buf) < msg_len:
            # If buffer isn't long enough, the message is incomplete
            return PARSE_RESULT.INCOMPLETE, buf, None

        msg = buf[0:msg_len]
        buf = buf[msg_len:]

        packet_type = msg[4]
        data = msg[6:6 + data_len]
        opt_data = msg[6 + data_len:6 + data_len + opt_len]

        # Check CRCs for header and data
        if msg[5] != crc8.calc(msg[1:5]):
            # Fail if doesn't match message
            logger.error('Header CRC error!')
            # Return CRC_MISMATCH
            return PARSE_RESULT.CRC_MISMATCH, buf, None
        if msg[6 + data_len + opt_len] != crc8.calc(msg[6:6 + data_len + opt_len]):
            # Fail if doesn't match message
            logger.error('Data CRC error!')
            # Return CRC_MISMATCH
            return PARSE_RESULT.CRC_MISMATCH, buf, None

        # If we got this far, everything went ok (?)
        if packet_type == PACKET.RADIO:
            p = RadioPacket(packet_type, data, opt_data)
        elif packet_type == PACKET.RESPONSE:
            p = ResponsePacket(packet_type, data, opt_data)
        else:
            p = Packet(packet_type, data, opt_data)

        return PARSE_RESULT.OK, buf, p

    def parse(self):
        ''' Parse data from Packet '''
        return self.parsed

    def parse_eep(self, func, type):
        ''' Parse EEP based on FUNC and TYPE '''
        provides, values = eep.get_values(self.rorg, func, type, self.bit_data)
        self.parsed.update(values)
        return list(provides)

    def build(self):
        ''' Build Packet for sending to EnOcean controller '''
        data_length = len(self.data)
        ords = [0x55, (data_length >> 8) & 0xFF, data_length & 0xFF, len(self.optional), int(self.type)]
        ords.append(crc8.calc(ords[1:5]))
        ords.extend(self.data)
        ords.extend(self.optional)
        ords.append(crc8.calc(ords[6:]))
        return ords


class RadioPacket(Packet):
    def parse(self):
        self.parsed['destination'] = self._combine_hex(self.optional[1:5])
        self.parsed['dBm'] = -self.optional[5]
        self.parsed['status'] = self.data[-1]
        self.parsed['sender'] = self._combine_hex(self.data[-5:-1])
        self.parsed['sender_hex'] = ':'.join([('%02X' % o) for o in self.data[-5:-1]])
        # Default to learn == True, as some devices don't have a learn button
        self.parsed['learn'] = True

        self.rorg = self.data[0]
        if self.rorg == RORG.RPS:
            self.bit_data = self._to_bitarray(self.data[1], 8)
        if self.rorg == RORG.BS1:
            self.bit_data = self._to_bitarray(self.data[1], 8)
            self.parsed['learn'] = not self.bit_data[-4]
        if self.rorg == RORG.BS4:
            self.bit_data = self._to_bitarray(self.data[1:5], 32)
            self.parsed['learn'] = not self.bit_data[-4]
            if self.parsed['learn']:
                self.parsed['contains_eep'] = self.bit_data[-7]
        return super(RadioPacket, self).parse()


class ResponsePacket(Packet):
    def parse(self):
        self.parsed['response'] = self.data[0]
        self.parsed['response_data'] = self.data[1:]
        return super(ResponsePacket, self).parse()


class EventPacket(Packet):
    def parse(self):
        self.parsed['event'] = self.data[0]
        self.parsed['event_data'] = self.data[1:]
        return super(EventPacket, self).parse()



# -*- encoding: utf-8 -*-
from __future__ import print_function, unicode_literals, division
import socket


def send_to_tcp_socket(host, port, packet):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.connect((host, port))
    s.send(str(bytearray(packet.build())))
    s.close()



# -*- encoding: utf-8 -*-
from __future__ import print_function, unicode_literals, division
import logging


def init_logging(level=logging.DEBUG):
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    logger = logging.getLogger('enocean')
    logger.setLevel(logging.DEBUG)

    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    ch.setFormatter(formatter)
    logger.addHandler(ch)



#!/usr/bin/env python
from distutils.core import setup

setup(
    name='enocean',
    version='0.15',
    description='EnOcean serial protocol implementation',
    author='Kimmo Huoman',
    author_email='kipenroskaposti@gmail.com',
    url='https://github.com/kipe/enocean',
    packages=[
        'enocean',
        'enocean.protocol',
        'enocean.communicators',
    ],
    scripts=[
        'examples/enocean_example.py',
    ],
    package_data={
        '': ['EEP_2.6.1.xml']
    },
    install_requires=[
        'enum34>=1.0',
        'pyserial>=2.7',
        'beautifulsoup4>=4.3.2',
    ])



# -*- encoding: utf-8 -*-
from __future__ import print_function, unicode_literals, division
from enum import IntEnum


# EnOceanSerialProtocol3.pdf / 12
class PACKET(IntEnum):
    RESERVED = 0x00
    RADIO = 0x01
    RESPONSE = 0x02
    RADIO_SUB_TEL = 0x03
    EVENT = 0x04
    COMMON_COMMAND = 0x05
    SMART_ACK_COMMAND = 0x06
    REMOTE_MAN_COMMAND = 0x07
    RADIO_MESSAGE = 0x09
    RADIO_ADVANCED = 0x0A


# EnOceanSerialProtocol3.pdf / 18
class RETURN_CODE(IntEnum):
    OK = 0x00
    ERROR = 0x01
    NOT_SUPPORTED = 0x02
    WRONG_PARAM = 0x04
    OPERATION_DENIED = 0x04


# EnOceanSerialProtocol3.pdf / 20
class EVENT_CODE(IntEnum):
    SA_RECLAIM_NOT_SUCCESFUL = 0x01
    SA_CONFIRM_LEARN = 0x02
    SA_LEARN_ACK = 0x03
    CO_READY = 0x04
    CO_EVENT_SECUREDEVICES = 0x05


# EnOcean_Equipment_Profiles_EEP_V2.61_public.pdf / 8
class RORG(IntEnum):
    UNDEFINED = 0x00
    RPS = 0xF6
    BS1 = 0xD5
    BS4 = 0xA5
    VLD = 0xD2
    MSC = 0xD1
    ADT = 0xA6
    SM_LRN_REQ = 0xC6
    SM_LRN_ANS = 0xC7
    SM_REC = 0xA7
    SYS_EX = 0xC5
    SEC = 0x30
    SEC_ENCAPS = 0x31


# Results for message parsing
class PARSE_RESULT(IntEnum):
    OK = 0x00
    INCOMPLETE = 0x01
    CRC_MISMATCH = 0x03
